#include "include/wl_def.h"
#include "include/32x.h"
#include "include/debug_32x.h"

#include <string.h>

byte *gfxbuf;
byte __attribute__((aligned(16))) gfxbuffer[320*200];

unsigned short currentFB;
volatile unsigned short dmaDone = 1;
int mousebuttons = 0;
int mousexaxis = 0;
int mouseyaxis = 0;

unsigned char gPalette[768];

extern void keyboard_handler(int code, int press);
extern boolean InternalKeyboard[NumCodes];
extern boolean MousePresent;

int main ( void )
{
    Debug32xInit(MARS_VDP_MODE_256);

    return WolfMain(0, NULL);
}

//void DisplayTextSplash(byte *text);

/*
==========================
=
= Quit
=
==========================
*/

void Quit(char *error)
{
    volatile unsigned short *palette = &MARS_CRAM;
    int i,j;

    Debug32xScreenPuts("\n");
    Debug32xScreenPuts(error);

    palette[255] = COLOR(31,31,31); // make sure text color is white

    j = 0;
    while (1)
    {
        palette[0] = COLOR(j, 0, 0);
        for (i=0;i<100000;) i++;
        j = (j+1) & 15;
    }

    memptr screen = NULL;

    if (!error || !*error) {
        CA_CacheGrChunk(ORDERSCREEN);
        screen = grsegs[ORDERSCREEN];
        //WriteConfig();
    } else if (error) {
        CA_CacheGrChunk(ERRORSCREEN);
        screen = grsegs[ERRORSCREEN];
    }

    ShutdownId();

    if (screen) {
        //DisplayTextSplash(screen);
    }

#if 0
    if (error && *error) {
        fprintf(stderr, "Quit: %s\n", error);
        exit(EXIT_FAILURE);
    }
#endif

    exit(EXIT_SUCCESS);
}

void VL_WaitVBL(int vbls)
{
    long last = get_TimeCount() + vbls;
    while (last > get_TimeCount()) ;
}

extern void fast_memcpy(void *dst, void *src, int len);
extern int SetSH2SR(int level);

void VW_UpdateScreen()
{
    byte *fb = (byte *)&MARS_FRAMEBUFFER;

    // wait for the framebuffer selection bit to take effect
    while ((MARS_VDP_FBCTL & MARS_VDP_FS) == (currentFB^1)) ;

    // copy screenbuffer to framebuffer
#if 0
    //memcpy(fb + 0x200 + 12 * 320, gfxbuf, 320*200);
    fast_memcpy(fb + 0x200 + 12 * 320, gfxbuf, 320*200/4);

    // Flip the framebuffer selection bit
    currentFB ^= 1;
    MARS_VDP_FBCTL = currentFB;
	dmaDone = 1;
#else
	dmaDone = 0;

    SH2_DMA_CHCR1; // read TE
    SH2_DMA_CHCR1 = 0; // clear TE

    // start DMA
    SH2_DMA_SAR1 = (uint32_t)gfxbuf;
    SH2_DMA_DAR1 = (uint32_t)(fb + 0x200 + 12 * 320);
    SH2_DMA_TCR1 = ((320*200 >> 4) << 2); // xfer count (4 * # of 16 byte units)
    SH2_DMA_CHCR1 = 0x5EE5; // dest incr, src incr, size 16B, auto req, cycle-steal, dual addr, intr enabled, clear TE, dma enabled
#endif
}

void main_dma1_handler(void)
{
    SH2_DMA_CHCR1; // read TE
    SH2_DMA_CHCR1 = 0; // clear TE

    // Flip the framebuffer selection bit
    currentFB ^= 1;
    MARS_VDP_FBCTL = currentFB;

	dmaDone = 1;
}

/*
=======================
=
= VL_Startup
=
=======================
*/

void VL_Startup()
{
    int i;
    volatile unsigned short *cram = &MARS_CRAM;
    volatile unsigned short *frameBuffer16 = &MARS_FRAMEBUFFER;

    // max value for 32X is 320x224
    vwidth = 320;
    vheight = 200;

    // Wait for the SH2 to gain access to the VDP
    while ((MARS_SYS_INTMSK & MARS_SH2_ACCESS_VDP) == 0) ;

    // Set 8-bit paletted mode, 224 lines
    MARS_VDP_DISPMODE = MARS_224_LINES | MARS_VDP_MODE_256;

    // init both framebuffers
    currentFB = 0;
    // Flip the framebuffer selection bit and wait for it to take effect
    MARS_VDP_FBCTL = currentFB ^ 1;
    while ((MARS_VDP_FBCTL & MARS_VDP_FS) == currentFB) ;
    currentFB ^= 1;
    // rewrite line table
    for (i=0; i<224; i++)
        frameBuffer16[i] = i*160 + 0x100; /* word offset of line */
    // Flip the framebuffer selection bit and wait for it to take effect
    MARS_VDP_FBCTL = currentFB ^ 1;
    while ((MARS_VDP_FBCTL & MARS_VDP_FS) == currentFB) ;
    currentFB ^= 1;
    // rewrite line table
    for (i=0; i<224; i++)
        frameBuffer16[i] = i*160 + 0x100; /* word offset of line */

    // clear the palette
    memset(gPalette, 0, 768);
    for (i=0; i<256; i++)
        cram[i] = COLOR(0, 0, 0);

    Debug32xSetFGColor(31,31,31);

    gfxbuf = (byte *)((unsigned int)gfxbuffer | 0x20000000); // uncached ptr
    vstride = 320;
    
    // init DMA
    SH2_DMA_SAR0 = 0;
    SH2_DMA_DAR0 = 0;
    SH2_DMA_TCR0 = 0;
    SH2_DMA_CHCR0 = 0;
    SH2_DMA_DRCR0 = 0;
    SH2_DMA_SAR1 = 0;
    SH2_DMA_DAR1 = 0;
    SH2_DMA_TCR1 = 0;
    SH2_DMA_CHCR1 = 0;
    SH2_DMA_DRCR1 = 0;
    SH2_DMA_DMAOR = 1; // enable DMA

    SH2_DMA_VCR1 = 72; // set exception vector for DMA channel 1
    SH2_INT_IPRA = (SH2_INT_IPRA & 0xF0FF) | 0x0F00; // set DMA INT to priority 15

	dmaDone = 1;
    SetSH2SR(2); // allow ints
}

/*
=======================
=
= VL_Shutdown
=
=======================
*/

void VL_Shutdown()
{
}

/* ======================================================================== */

/*
=================
=
= VL_SetPalette
=
=================
*/

void VL_SetPalette(const byte *palette)
{
    int i;
    volatile unsigned short *cram = &MARS_CRAM;

    //VL_WaitVBL(1);

    for (i = 0; i < 256; i++)
    {
        gPalette[i*3+0] = palette[i*3+0] << 2;
        gPalette[i*3+1] = palette[i*3+1] << 2;
        gPalette[i*3+2] = palette[i*3+2] << 2;

        // 6-bit / gun color => 5-bit / gun cram entry
        cram[i] = COLOR(palette[i*3+0]>>1,palette[i*3+1]>>1,palette[i*3+2]>>1);
    }

    //Debug32xSetFGColor(31,31,31);

    // Should also do a flip here.
    void VW_UpdateScreen();
}

/*
=================
=
= VL_GetPalette
=
=================
*/

void VL_GetPalette(byte *palette)
{
    int i;
    for (i=0;i<256;i++)
    {
        palette[i*3+0] = gPalette[i*3+0] >> 2;
        palette[i*3+1] = gPalette[i*3+1] >> 2;
        palette[i*3+2] = gPalette[i*3+2] >> 2;
    }
}

/*
=================
=
= INL_Update
=
=================
*/

static int mx = 0;
static int my = 0;

void INL_Update()
{
    static unsigned short buttons = 0;
    unsigned short new_buttons, curr_buttons;

    // MARS_SYS_COMM10 holds the current button values: - - - - M X Y Z S A C B R L D U
    curr_buttons = MARS_SYS_COMM8;
    if ((curr_buttons & SEGA_CTRL_TYPE) == SEGA_CTRL_NONE)
        curr_buttons = MARS_SYS_COMM10;     // if no pad 1, try using pad 2

    new_buttons = curr_buttons ^ buttons; // set if button changed
    buttons = curr_buttons;

    // we use mouse to move, so this doesn't do anything, so don't use it (unless using mouse)
    if (MousePresent) {
    if (new_buttons & SEGA_CTRL_A)
    {
        if (!(buttons & SEGA_CTRL_A))
        {
            // A just released
            keyboard_handler(sc_RShift, 0); // run not pressed
        }
        else
        {
            // A just pressed
            if (!(buttons & SEGA_CTRL_START))
            {
                keyboard_handler(sc_RShift, 1); // run pressed
            }
        }
    }
    }

    if (new_buttons & SEGA_CTRL_B)
    {
        if (!(buttons & SEGA_CTRL_B))
        {
            // B just released
            keyboard_handler(sc_Control, 0); // FIRE not pressed
            keyboard_handler(sc_Enter, 0); // ENTER not pressed
            //keyboard_handler(sc_Y, 0); // Y not pressed
        }
        else
        {
            // B just pressed
            if (!(buttons & SEGA_CTRL_START))
            {
                keyboard_handler(sc_Control, 1); // FIRE pressed
                keyboard_handler(sc_Enter, 1); // ENTER pressed
                //keyboard_handler(sc_Y, 1); // Y pressed
            }
        }
    }
    if (new_buttons & SEGA_CTRL_C)
    {
        if (!(buttons & SEGA_CTRL_C))
        {
            // C just released
            keyboard_handler(sc_Alt, 0); // strafe not pressed
            keyboard_handler(sc_Space, 0); // open/operate not pressed
            keyboard_handler(sc_Escape, 0); // MENU not pressed
        }
        else
        {
            // C just pressed
            if (!(buttons & SEGA_CTRL_START))
            {
                keyboard_handler(sc_Alt, 1); // strafe pressed
                keyboard_handler(sc_Space, 1); // open/operate pressed
            }
            else
            {
                keyboard_handler(sc_Escape, 1); // MENU pressed
            }
        }
    }
    if (buttons & SEGA_CTRL_START && new_buttons & SEGA_CTRL_X)
    {
        if (!(buttons & SEGA_CTRL_X))
        {
            // X just released
            keyboard_handler(sc_F9, 0); // F9 not pressed
        }
        else
        {
            // START + X just pressed
            keyboard_handler(sc_F9, 1); // F9 (Quick Load) pressed
        }
    }
    if (buttons & SEGA_CTRL_START && new_buttons & SEGA_CTRL_Y)
    {
        if (!(buttons & SEGA_CTRL_Y))
        {
            // Y just released
            keyboard_handler(sc_F8, 0); // F8 not pressed
        }
        else
        {
            // START + Y just pressed
            keyboard_handler(sc_F8, 1); // F8 (Quick Save) pressed
        }
    }
    if (new_buttons & SEGA_CTRL_Z)
    {
        if (!(buttons & SEGA_CTRL_Z))
        {
            // Z just released
            keyboard_handler(sc_Escape, 0); // MENU not pressed
        }
        else
        {
            // Z just pressed
            keyboard_handler(sc_Escape, 1); // MENU pressed
        }
    }

    if (new_buttons & SEGA_CTRL_UP)
    {
        if (!(buttons & SEGA_CTRL_UP))
        {
            // UP just released
            my = 0;
            keyboard_handler(sc_UpArrow, 0);
            keyboard_handler(sc_1, 0); // weapon 1 not pressed
            keyboard_handler(sc_Tab, 0); // release god-mode keys
            keyboard_handler(sc_G, 0);
            keyboard_handler(sc_F10, 0);
        }
        else
        {
            // UP just pressed
            if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_A) || buttons & SEGA_CTRL_X)
            {
                keyboard_handler(sc_1, 1); // weapon 1 pressed
            }
            else if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_B) || buttons & SEGA_CTRL_Y)
            {
                keyboard_handler(sc_Tab, 1);
                keyboard_handler(sc_G, 1);
                keyboard_handler(sc_F10, 1); // toggle god-mode
            }
        }
    }
    if (new_buttons & SEGA_CTRL_RIGHT)
    {
        if (!(buttons & SEGA_CTRL_RIGHT))
        {
            // RIGHT just released
            mx = 0;
            keyboard_handler(sc_RightArrow, 0);
            keyboard_handler(sc_2, 0); // weapon 2 not pressed
        }
        else
        {
            // RIGHT just pressed
            if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_A) || buttons & SEGA_CTRL_X)
            {
                keyboard_handler(sc_2, 1); // weapon 2 pressed
            }
            else if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_B) || buttons & SEGA_CTRL_Y)
            {
                viewsize = (viewsize < 20) ? viewsize + 2 : 20; // increase view size
                NewViewSize(viewsize);
                DrawPlayBorder();
            }
        }
    }
    if (new_buttons & SEGA_CTRL_DOWN)
    {
        if (!(buttons & SEGA_CTRL_DOWN))
        {
            // DOWN just released
            my = 0;
            keyboard_handler(sc_DownArrow, 0);
            keyboard_handler(sc_3, 0); // weapon 3 not pressed
            keyboard_handler(sc_M, 0); // release health/ammo/keys/weapons keys
            keyboard_handler(sc_I, 0);
            keyboard_handler(sc_L, 0);
        }
        else
        {
            // DOWN just pressed
            if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_A) || buttons & SEGA_CTRL_X)
            {
                keyboard_handler(sc_3, 1); // weapon 3 pressed
            }
            else if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_B) || buttons & SEGA_CTRL_Y)
            {
                keyboard_handler(sc_M, 1);
                keyboard_handler(sc_I, 1);
                keyboard_handler(sc_L, 1); // Full Health, Ammo, Keys, Weapons, Zero score
            }
        }
    }
    if (new_buttons & SEGA_CTRL_LEFT)
    {
        if (!(buttons & SEGA_CTRL_LEFT))
        {
            // LEFT just released
            mx = 0;
            keyboard_handler(sc_LeftArrow, 0);
            keyboard_handler(sc_4, 0); // weapon 4 not pressed
        }
        else
        {
            // LEFT just pressed
            if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_A) || buttons & SEGA_CTRL_X)
            {
                keyboard_handler(sc_4, 1); // weapon 4 pressed
            }
            else if ((buttons & SEGA_CTRL_START && buttons & SEGA_CTRL_B) || buttons & SEGA_CTRL_Y)
            {
                viewsize = (viewsize > 10) ? viewsize - 2 : 10; // decrease view size
                NewViewSize(viewsize);
                DrawPlayBorder();
            }
        }
    }

    // do movement
    #define OTHER_BTNS (SEGA_CTRL_START | SEGA_CTRL_X | SEGA_CTRL_Y)
    if ((buttons & SEGA_CTRL_UP) && !(buttons & OTHER_BTNS))
    {
        if (MousePresent)
        {
            keyboard_handler(sc_UpArrow, 1);
        }
        else
        {
            if (buttons & SEGA_CTRL_A)
                my = -40; // run
            else
                my = -20;
        }
    }
    if ((buttons & SEGA_CTRL_RIGHT) && !(buttons & OTHER_BTNS))
    {
        if (MousePresent)
        {
            keyboard_handler(sc_RightArrow, 1);
        }
        else
        {
            if (buttons & SEGA_CTRL_A)
                mx = 40; // run
            else
                mx = 20;
        }
    }
    if ((buttons & SEGA_CTRL_DOWN) && !(buttons & OTHER_BTNS))
    {
        if (MousePresent)
        {
            keyboard_handler(sc_DownArrow, 1);
        }
        else
        {
            if (buttons & SEGA_CTRL_A)
                my = 40; // run
            else
                my = 20;
        }
    }
    if ((buttons & SEGA_CTRL_LEFT) && !(buttons & OTHER_BTNS))
    {
        if (MousePresent)
        {
            keyboard_handler(sc_LeftArrow, 1);
        }
        else
        {
            if (buttons & SEGA_CTRL_A)
                mx = -40; // run
            else
                mx = -20;
        }
    }

    // misc buttons
    if (new_buttons & SEGA_CTRL_B && !StartGame)
    {
        if (!(buttons & SEGA_CTRL_B))
        {
            // B just released
            keyboard_handler(sc_Enter, 0); // Enter not pressed
        }
        else
        {
            // B just pressed
            keyboard_handler(sc_Enter, 1); // Enter pressed
        }
    }
    if (new_buttons & SEGA_CTRL_C && !StartGame)
    {
        if (!(buttons & SEGA_CTRL_C))
        {
            // C just released
            keyboard_handler(sc_Escape, 0); // MENU not pressed
        }
        else
        {
            // C just pressed
            keyboard_handler(sc_Escape, 1); // MENU pressed
        }
    }
    if (new_buttons & SEGA_CTRL_MODE)
    {
        if (!(buttons & SEGA_CTRL_MODE))
        {
            // MODE just released
            keyboard_handler(sc_A, 0); // Automap not pressed
        }
        else
        {
            // MODE just pressed
            keyboard_handler(sc_A, 1); // Automap pressed
        }
    }

}

static int get_mouse(int port)
{
    unsigned int mouse1, mouse2;
    while (MARS_SYS_COMM0) ; // wait until 68000 has responded to any earlier requests
    MARS_SYS_COMM0 = 0x0500|port; // tells 68000 to read mouse
    while (MARS_SYS_COMM0 == (0x0500|port)) ; // wait for mouse value
    mouse1 = MARS_SYS_COMM0;
    mouse2 = MARS_SYS_COMM2;
    MARS_SYS_COMM0 = 0; // tells 68000 we got the mouse value
    return (int)((mouse1 << 16) | mouse2);
}

void IN_GetMouseDelta(int *dx, int *dy)
{
    if (MousePresent)
    {
        if (dx)
            *dx = mousexaxis;
        if (dy)
            *dy = -mouseyaxis;
        mousexaxis = 0;
        mouseyaxis = 0;
        return;
    }
    if (dx)
        *dx = mx;
    if (dy)
        *dy = my;
}

byte IN_MouseButtons()
{
    static unsigned int lastmousetick = -1;

    if (MousePresent)
    {
        if (lastmousetick != MARS_SYS_COMM12)
        {
            int val = get_mouse(0);
            if (val != -2)
            {
                int mx, my;
                mousebuttons = (val >> 16) & 15;
                mx = (val >> 8) & 0xFF;
                // check sign
                if (val & 0x00100000)
                    mx |= 0xFFFFFF00;
                // check overflow
                if (val & 0x00400000)
                    mx = (val & 0x00100000) ? -256 : 256;
                mousexaxis += mx;
                my = val & 0xFF;
                // check sign
                if (val & 0x00200000)
                    my |= 0xFFFFFF00;
                // check overflow
                if (val & 0x00800000)
                    my = (val & 0x00200000) ? -256 : 256;
                mouseyaxis += my;
            }
            lastmousetick = MARS_SYS_COMM12;
        }
    }
    return (mousebuttons & 7);
}
