#include "include/wl_def.h"
#include "include/32x.h"
#include "include/debug_32x.h"

/*
=============================================================================

						   WOLFENSTEIN 3-D

					  An Id Software production

						   by John Carmack

=============================================================================
*/

#define FOCALLENGTH     0x5800		/* in global coordinates */

char str[80], str2[20];

int viewwidth, viewheight;
int viewwidthwin, viewheightwin; /* for borders */
int xoffset, yoffset;
int vwidth, vheight, vstride; /* size of screen */
int viewsize;

int centerx;
int shootdelta;			/* pixels away from centerx a target can be */

boolean startgame,loadedgame;
int mouseadjustment;

long frameon;
long lasttimecount;
fixed viewsin, viewcos;
fixed viewx, viewy;		/* the focal point */
int pixelangle[MAXVIEWWIDTH];
long finetangent[FINEANGLES/4];
int horizwall[MAXWALLTILES], vertwall[MAXWALLTILES];

char configname[13] = "config.";

fixed sintable[ANGLES+ANGLES/4+1], *costable = sintable+(ANGLES/4);

int _argc;
char **_argv;

// comment out the next line if you don't wish to use compression
#define USE_LZ77

#ifdef USE_LZ77
/*
 * LZ77 support code - derived from lz.c by Marcus Geelnard
 * 
 */

/*************************************************************************
* Copyright (c) 2003-2006 Marcus Geelnard
*
* This software is provided 'as-is', without any express or implied
* warranty. In no event will the authors be held liable for any damages
* arising from the use of this software.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would
*    be appreciated but is not required.
*
* 2. Altered source versions must be plainly marked as such, and must not
*    be misrepresented as being the original software.
*
* 3. This notice may not be removed or altered from any source
*    distribution.
*
* Marcus Geelnard
* marcus.geelnard at home.se
*************************************************************************/

// The max offset controls the compression level (and hence speed). Set
//   to 32768, compressing a save can take over 2 minutes. At 8192, it
//   takes just under a minute. At 2048, it takes just slightly more than
//   15 seconds.
//#define LZ_MAX_OFFSET 32768
//#define LZ_MAX_OFFSET 8192
#define LZ_MAX_OFFSET 2048

unsigned char cmpbuf[32768-64-4];

static unsigned int _LZ_StringCompare( unsigned char * str1, unsigned char * str2, unsigned int minlen, unsigned int maxlen )
{
    unsigned int len;

    for( len = minlen; (len < maxlen) && (str1[len] == str2[len]); ++ len );

    return len;
}

static int _LZ_WriteVarSize( unsigned int x, unsigned char * buf )
{
    unsigned int y;
    int num_bytes, i, b;

    /* Determine number of bytes needed to store the number x */
    y = x >> 3;
    for( num_bytes = 5; num_bytes >= 2; -- num_bytes )
    {
        if( y & 0xfe000000 ) break;
        y <<= 7;
    }

    /* Write all bytes, seven bits in each, with 8:th bit set for all */
    /* but the last byte. */
    for( i = num_bytes-1; i >= 0; -- i )
    {
        b = (x >> (i*7)) & 0x0000007f;
        if( i > 0 )
        {
            b |= 0x00000080;
        }
        *buf ++ = (unsigned char) b;
    }

    /* Return number of bytes written */
    return num_bytes;
}

static int _LZ_ReadVarSize( unsigned int * x, unsigned char * buf )
{
    unsigned int y, b, num_bytes;

    /* Read complete value (stop when byte contains zero in 8:th bit) */
    y = 0;
    num_bytes = 0;
    do
    {
        b = (unsigned int) (*buf ++);
        y = (y << 7) | (b & 0x0000007f);
        ++ num_bytes;
    }
    while( b & 0x00000080 );

    /* Store value in x */
    *x = y;

    /* Return number of bytes read */
    return num_bytes;
}

int Compress( unsigned char *in, unsigned char *out, unsigned int insize )
{
    unsigned char marker, symbol;
    unsigned int  inpos, outpos, bytesleft, i;
    unsigned int  maxoffset, offset, bestoffset;
    unsigned int  maxlength, length, bestlength;
    unsigned int  histogram[ 256 ];
    unsigned char *ptr1, *ptr2;

    /* Do we have anything to compress? */
    if( insize < 1 )
    {
        return 0;
    }

    /* Create histogram */
    for( i = 0; i < 256; ++ i )
    {
        histogram[ i ] = 0;
    }
    for( i = 0; i < insize; ++ i )
    {
        ++ histogram[ in[ i ] ];
    }

    /* Find the least common byte, and use it as the marker symbol */
    marker = 0;
    for( i = 1; i < 256; ++ i )
    {
        if( histogram[ i ] < histogram[ marker ] )
        {
            marker = i;
        }
    }

    /* Remember the marker symbol for the decoder */
    out[ 0 ] = marker;

    /* Start of compression */
    inpos = 0;
    outpos = 1;

    /* Main compression loop */
    bytesleft = insize;
    do
    {
        /* Determine most distant position */
        if( inpos > LZ_MAX_OFFSET ) maxoffset = LZ_MAX_OFFSET;
        else                        maxoffset = inpos;

        /* Get pointer to current position */
        ptr1 = &in[ inpos ];

        /* Search history window for maximum length string match */
        bestlength = 3;
        bestoffset = 0;
        for( offset = 3; offset <= maxoffset; ++ offset )
        {
            /* Get pointer to candidate string */
            ptr2 = &ptr1[ -(int)offset ];

            /* Quickly determine if this is a candidate (for speed) */
            if( (ptr1[ 0 ] == ptr2[ 0 ]) &&
                (ptr1[ bestlength ] == ptr2[ bestlength ]) )
            {
                /* Determine maximum length for this offset */
                maxlength = (bytesleft < offset ? bytesleft : offset);

                /* Count maximum length match at this offset */
                length = _LZ_StringCompare( ptr1, ptr2, 0, maxlength );

                /* Better match than any previous match? */
                if( length > bestlength )
                {
                    bestlength = length;
                    bestoffset = offset;
                }
            }
        }

        /* Was there a good enough match? */
        if( (bestlength >= 8) ||
            ((bestlength == 4) && (bestoffset <= 0x0000007f)) ||
            ((bestlength == 5) && (bestoffset <= 0x00003fff)) ||
            ((bestlength == 6) && (bestoffset <= 0x001fffff)) ||
            ((bestlength == 7) && (bestoffset <= 0x0fffffff)) )
        {
            out[ outpos ++ ] = (unsigned char) marker;
            outpos += _LZ_WriteVarSize( bestlength, &out[ outpos ] );
            outpos += _LZ_WriteVarSize( bestoffset, &out[ outpos ] );
            inpos += bestlength;
            bytesleft -= bestlength;
        }
        else
        {
            /* Output single byte (or two bytes if marker byte) */
            symbol = in[ inpos ++ ];
            out[ outpos ++ ] = symbol;
            if( symbol == marker )
            {
                out[ outpos ++ ] = 0;
            }
            -- bytesleft;
        }
    }
    while( bytesleft > 3 );

    /* Dump remaining bytes, if any */
    while( inpos < insize )
    {
        if( in[ inpos ] == marker )
        {
            out[ outpos ++ ] = marker;
            out[ outpos ++ ] = 0;
        }
        else
        {
            out[ outpos ++ ] = in[ inpos ];
        }
        ++ inpos;
    }

    return outpos;
}

void Decompress( unsigned char *in, unsigned char *out, unsigned int insize )
{
    unsigned char marker, symbol;
    unsigned int  i, inpos, outpos, length, offset;

    /* Do we have anything to uncompress? */
    if( insize < 1 )
    {
        return;
    }

    /* Get marker symbol from input stream */
    marker = in[ 0 ];
    inpos = 1;

    /* Main decompression loop */
    outpos = 0;
    do
    {
        symbol = in[ inpos ++ ];
        if( symbol == marker )
        {
            /* We had a marker byte */
            if( in[ inpos ] == 0 )
            {
                /* It was a single occurrence of the marker byte */
                out[ outpos ++ ] = marker;
                ++ inpos;
            }
            else
            {
                /* Extract true length and offset */
                inpos += _LZ_ReadVarSize( &length, &in[ inpos ] );
                inpos += _LZ_ReadVarSize( &offset, &in[ inpos ] );

                /* Copy corresponding data from history window */
                for( i = 0; i < length; ++ i )
                {
                    out[ outpos ] = out[ outpos - offset ];
                    ++ outpos;
                }
            }
        }
        else
        {
            /* No marker, plain copy */
            out[ outpos ++ ] = symbol;
        }
    }
    while( inpos < insize );
}
#endif

/*
 * SRAM support code
 *
 */

#define NUM_SLOTS 2

static int sram_offset[NUM_SLOTS];

int SeekSRAM(int slot, int offset, int mode)
{
	if ((slot >= NUM_SLOTS) || (slot < 0))
		return 0;

	switch (mode) {
	case SEEK_SET:
		sram_offset[slot] = offset;
		break;
	case SEEK_CUR:
		sram_offset[slot] += offset;
		break;
	case SEEK_END:
		sram_offset[slot] = 16383 + offset;
		break;
	}
	if (sram_offset[slot] < 0)
		sram_offset[slot] = 0;
	else if (sram_offset[slot] > 16383)
		sram_offset[slot] = 16383;

	return sram_offset[slot];
}

void ReadSRAM(int slot, byte *buffer, int size)
{
	byte *ptr = buffer;
	int len = size;

	if ((slot >= NUM_SLOTS) || (slot < 0))
		return 0;

	if ((sram_offset[slot] + len) > 16383)
		len = 16383 - sram_offset[slot];

#if 1
extern volatile unsigned short dmaDone;

	while (!dmaDone) ; // wait until screen has been copied to frame buffer
#endif

	while (MARS_SYS_COMM0) ;
	while (len > 0) {
		MARS_SYS_COMM2 = sram_offset[slot] + (slot * 16384);
		MARS_SYS_COMM0 = 0x0100;	/* Read SRAM */
		while (MARS_SYS_COMM0) ;
		*ptr++ = MARS_SYS_COMM2 & 0x00FF;
		sram_offset[slot]++;
		len--;
	}
}

void WriteSRAM(int slot, byte *buffer, int size)
{
	byte *ptr = buffer;
	int len = size;

	if ((slot >= NUM_SLOTS) || (slot < 0))
		return 0;

	if ((sram_offset[slot] + len) > 16383)
		len = 16383 - sram_offset[slot];

#if 1
extern volatile unsigned short dmaDone;

	while (!dmaDone) ; // wait until screen has been copied to frame buffer
#endif

	while (MARS_SYS_COMM0) ;
	while (len > 0) {
		MARS_SYS_COMM2 = sram_offset[slot] + (slot * 16384);
		MARS_SYS_COMM0 = 0x0200 | *ptr++;	/* Write SRAM */
		while (MARS_SYS_COMM0) ;
		sram_offset[slot]++;
		len--;
	}
}


/*
========================
=
= FixedByFrac (FixedMul)
=
= multiply two 16/16 bit, 2's complement fixed point numbers
=
========================
*/

fixed FixedByFrac(fixed a, fixed b)
{
    fixed c;
    asm("mov %0,r1"::"r"(a):"r1");
    asm("mov %0,r2"::"r"(b):"r2");
    asm("dmuls.l r1,r2");
    asm("sts mach,r1");
    asm("sts macl,r2");
    asm("xtrct r1,r2");
    asm("mov r2,%0":"=g"(c));
    return c;
}

/*
=====================
=
= CalcTics
=
=====================
*/

void CalcTics()
{
	int newtime;
	int ticcount;

	if (demoplayback || demorecord)
		ticcount = DEMOTICS - 1; /* [70/4] 17.5 Hz */
	else
		ticcount = 0 + 1; /* [70/2] 35 Hz */

	do {
		newtime = get_TimeCount();
		tics = newtime - lasttimecount;
	} while (tics <= ticcount);

	lasttimecount = newtime;

	if (demoplayback || demorecord)
		tics = DEMOTICS;
	else if (tics > MAXTICS)
		tics = MAXTICS;
}

/* ======================================================================== */

static void DiskFlopAnim(int x, int y)
{
	static char which = 0;

	if (!x && !y)
		return;

	VWB_DrawPic(x, y, C_DISKLOADING1PIC+which);
	VW_UpdateScreen();

	which ^= 1;
}

#if 0
static int32_t CalcFileChecksum(int fd, int len)
{
	int32_t cs;
	int i;
	int8_t c1, c2;

	c1 = ReadInt8(fd);
	cs = 0;
	for (i = 0; i < len - 1; i++) {
		c2 = ReadInt8(fd);
		cs += c1 ^ c2;
		c1 = c2;
	}
	return cs;
}
#endif

int WriteConfig()
{
#if 0
	int i;
	int fd;
	int32_t cs;

	fd = OpenWrite(configname);

	if (fd != -1) {
		WriteBytes(fd, (byte *)GAMEHDR, 8);	/* WOLF3D, 0, 0 */
		WriteBytes(fd, (byte *)CFGTYPE, 4);	/* CFG, 0 */
		WriteInt32(fd, 0xFFFFFFFF);		/* Version (integer) */
		WriteBytes(fd, (byte *)GAMETYPE, 4);	/* XXX, 0 */
		WriteInt32(fd, time(NULL));		/* Time */
		WriteInt32(fd, 0x00000000);		/* Padding */
		WriteInt32(fd, 0x00000000);		/* Checksum (placeholder) */

		for (i = 0; i < 7; i++) { /* MaxScores = 7 */
			WriteBytes(fd, (byte *)Scores[i].name, 58);
			WriteInt32(fd, Scores[i].score);
			WriteInt32(fd, Scores[i].completed);
			WriteInt32(fd, Scores[i].episode);
		}

		WriteInt32(fd, viewsize);


        WriteInt32(fd, viewsize);

		/* sound config, etc. (to be done) */
		WriteInt32(fd, SoundMode); /* padding */
		WriteInt32(fd, MusicMode); /* padding */
		WriteInt32(fd, DigiMode); /* padding */
		WriteInt32(fd, SD_GetVolume()); /* padding */
		WriteInt32(fd, 0); /* padding */
		WriteInt32(fd, 0); /* padding */
		WriteInt32(fd, 0); /* padding */
		WriteInt32(fd, 0); /* padding */

		fd = OpenRead(configname);
		ReadSeek(fd, 32, SEEK_SET);
		cs = CalcFileChecksum(fd, ReadLength(fd) - 32);
		CloseRead(fd);

		fd = OpenWriteAppend(configname);
		WriteSeek(fd, 28, SEEK_SET);
		WriteInt32(fd, cs);

		CloseWrite(fd);
	}
#endif

	return 0;
}

static void SetDefaults()
{
	viewsize = 16; // 4 to 40, based on units of 16
	SD_SetSoundMode(sdm_AdLib);
	SD_SetDigiDevice(sds_SoundBlaster);
	SD_SetMusicMode(smm_AdLib);
	SD_SetVolume(10);
}

int ReadConfig()
{
#if 0
	int fd, configokay;
	char buf[8];
	int32_t version, v;
	int i;

	configokay = 0;

	fd = OpenRead(configname);

	if (fd != -1) {
		SetDefaults();

		ReadBytes(fd, (byte *)buf, 8);
		if (strncmp(buf, GAMEHDR, 8))
			goto configend;

		ReadBytes(fd, (byte *)buf, 4);
		if (strncmp(buf, CFGTYPE, 4))
			goto configend;

		version = ReadInt32(fd);
		if (version != 0xFFFFFFFF && version != 0x00000000)
			goto configend;

		ReadBytes(fd, (byte *)buf, 4);
		if (strncmp(buf, GAMETYPE, 4))
			goto configend;

		ReadInt32(fd);	/* skip over time */
		ReadInt32(fd);	/* skip over padding */

		v = ReadInt32(fd);	/* get checksum */
		if (v != CalcFileChecksum(fd, ReadLength(fd) - 32))
			goto configend;

		ReadSeek(fd, 32, SEEK_SET);

		for (i = 0; i < 7; i++) { /* MaxScores = 7 */
			ReadBytes(fd, (byte *)Scores[i].name, 58);
			Scores[i].score = ReadInt32(fd);
			Scores[i].completed = ReadInt32(fd);
			Scores[i].episode = ReadInt32(fd);
		}

		viewsize = ReadInt32(fd);

		/* load the new data */
		if (version == 0x00000000) {
			/* sound config, etc. */
			SD_SetSoundMode(ReadInt32(fd)); /* padding */
			SD_SetMusicMode(ReadInt32(fd)); /* padding */
			SD_SetDigiDevice(ReadInt32(fd)); /* padding */
			SD_SetVolume(ReadInt32(fd)); /* padding */
			ReadInt32(fd); /* padding */
			ReadInt32(fd); /* padding */
			ReadInt32(fd); /* padding */
			ReadInt32(fd); /* padding */

			/* direction keys */
			for (i = 0; i < 4; i++) {
				dirscan[i] = ReadInt32(fd);
			}

			/* other game keys */
			for (i = 0; i < 8; i++) { /* NUMBUTTONS = 8 */
				buttonscan[i] = ReadInt32(fd);
			}

			/* mouse enabled */
			mouseenabled = ReadInt8(fd);

			/* mouse buttons */
			for (i = 0; i < 4; i++) {
				buttonmouse[i] = ReadInt32(fd);
			}

			/* mouse adjustment */
			mouseadjustment = ReadInt32(fd);

			/* unimplemented joystick */
			v = ReadInt32(fd);
			if (v != 0xFFFFFFFF) {
			}
		}

#ifdef UPLOAD
		MainMenu[readthis].active = 1;
		MainItems.curpos = 0;
#endif

		configokay = 1;
	}

configend:

	if (fd != -1)
		CloseRead(fd);

	if (!configokay) {
		//printf("Config: Setting defaults..\n");
		SetDefaults();
	}
#else
    SetDefaults();
#endif

	mouseenabled = true;

	joystickenabled = false;
	joypadenabled = false;
	joystickport = 0;

	mouseadjustment = 9;

	return 0;
}

/* ======================================================================== */

#if 1
long DoChecksum(byte *source, int size, long checksum)
{
	int i;

	for (i=0; i<size-1; i++)
		checksum += source[i] ^ source[i+1];

	return checksum;
}
#endif

int SaveTheGame(char *fn, char *tag, int x, int y)
{
#ifndef USE_LZ77
	int fd;
	int i;
	objtype *ob, nullobj;
	long checksum, v;

	fd = fn[7] - '0'; // filename was made like this - name[7] = which+'0';
	if ((fd >= NUM_SLOTS) || (fd < 0))
		return -1;

	SeekSRAM(fd, 0, SEEK_SET);

	v = 0xFFFFFFFF;

	// write savegame header
	WriteSRAM(fd, (byte *)GAMEHDR, 8);
	WriteSRAM(fd, (byte *)SAVTYPE, 4);
	WriteSRAM(fd, (byte *)&v, 4); /* write version */
	WriteSRAM(fd, (byte *)GAMETYPE, 4);

	v = 0;

	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)tag, 32); /* write savegame name */

	checksum = 0;

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)&gamestate, sizeof(gamestate));
	checksum = DoChecksum((byte*)&gamestate, sizeof(gamestate), checksum);

	DiskFlopAnim(x,y);

#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		WriteSRAM(fd, (byte *)&LevelRatios[i], sizeof(LRstruct));
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap, sizeof(tilemap), checksum);

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	WriteSRAM(fd, (byte *)areaconnect, sizeof(areaconnect));
	WriteSRAM(fd, (byte *)areabyplayer, sizeof(areabyplayer));

	for (ob = player; ob; ob = ob->next)
	{
		DiskFlopAnim(x,y);

		WriteSRAM(fd, (byte *)ob, sizeof(*ob));
	}
	nullobj.active = ac_badobject;          // end of file marker

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)&nullobj, sizeof(nullobj));

	DiskFlopAnim(x,y);

	i = laststatobj - statobjlist;
	WriteSRAM(fd, (byte *)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	DiskFlopAnim(x,y);

	WriteSRAM(fd, (byte *)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);

	WriteSRAM(fd, (byte *)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);

	WriteSRAM(fd, (byte *)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);

	WriteSRAM(fd, (byte *)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir, sizeof(pwalldir),checksum);

	WriteSRAM(fd, (byte *)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	DiskFlopAnim(x,y);

//	WriteSRAM(fd, (byte *)&Auto_Map, sizeof(automap_t));
//	checksum = DoChecksum((byte*)&Auto_Map, sizeof(automap_t), checksum);

	// WRITE OUT CHECKSUM

	WriteSRAM(fd, (byte *)&checksum, sizeof(checksum));
#else
	int fd;
	int i;
	objtype *ob, nullobj;
	long checksum, v;
	byte *umem, *cmem;
	int usize, csize;

#define WriteRAM(m, s) { memcpy(umem, m, s); umem += s; usize += s; }

	umem = gfxbuf;
	usize = 0;

	VL_WaitVBL(5);

	fd = fn[7] - '0'; // filename was made like this - name[7] = which+'0';
	if ((fd >= NUM_SLOTS) || (fd < 0))
		return -1;

	SeekSRAM(fd, 0, SEEK_SET);

	v = 0xFFFFFFFF;

	// write savegame header
	WriteSRAM(fd, (byte *)GAMEHDR, 8);
	WriteSRAM(fd, (byte *)SAVTYPE, 4);
	WriteSRAM(fd, (byte *)&v, 4); /* write version */
	WriteSRAM(fd, (byte *)GAMETYPE, 4);

	v = 0;

	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)&v, 4);
	WriteSRAM(fd, (byte *)tag, 32); /* write savegame name */

	checksum = 0;

	WriteRAM((byte *)&gamestate, sizeof(gamestate));
	checksum = DoChecksum((byte*)&gamestate, sizeof(gamestate), checksum);

#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		WriteRAM((byte *)&LevelRatios[i], sizeof(LRstruct));
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	WriteRAM((byte *)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap, sizeof(tilemap), checksum);

	WriteRAM((byte *)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	WriteRAM((byte *)areaconnect, sizeof(areaconnect));
	WriteRAM((byte *)areabyplayer, sizeof(areabyplayer));

	for (ob = player; ob; ob = ob->next)
	{
		WriteRAM((byte *)ob, sizeof(*ob));
	}
	nullobj.active = ac_badobject;          // end of file marker

	WriteRAM((byte *)&nullobj, sizeof(nullobj));

	i = laststatobj - statobjlist;
	WriteRAM((byte *)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);

	WriteRAM((byte *)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	WriteRAM((byte *)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);

	WriteRAM((byte *)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	WriteRAM((byte *)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);

	WriteRAM((byte *)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);

	WriteRAM((byte *)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);

	WriteRAM((byte *)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir, sizeof(pwalldir),checksum);

	WriteRAM((byte *)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	WriteRAM((byte *)&Auto_Map, sizeof(automap_t));
	checksum = DoChecksum((byte*)&Auto_Map, sizeof(automap_t), checksum);

	WriteRAM((byte *)&checksum, sizeof(checksum));

	umem = gfxbuf;
	cmem = cmpbuf;

	csize = Compress(umem, cmem, usize);
	if (csize && (csize <= (16384-64-4)))
	{
		SeekSRAM(fd, 64, SEEK_SET);
		WriteSRAM(fd, (byte*)&csize, 4); // save compressed size
		WriteSRAM(fd, cmem, csize); // save compressed memory
	}
	else
	{
		csize = 0;
		SeekSRAM(fd, 64, SEEK_SET);
		WriteSRAM(fd, (byte*)&csize, 4); // csize = 0 = no save
		return -1;
	}
#endif

	return 0;
}

int ReadSaveTag(int slot, char *tag)
{
	char buf[8];
	int fd;
	int32_t v;

	fd = slot;
	if ((fd >= NUM_SLOTS) || (fd < 0))
		return -1;

	SeekSRAM(fd, 0, SEEK_SET);

	ReadSRAM(fd, (byte *)buf, 8);
	if (strncmp(buf, GAMEHDR, 8))
		return -1;

	ReadSRAM(fd, (byte *)buf, 4);
	if (strncmp(buf, SAVTYPE, 4))
		return -1;

	ReadSRAM(fd, (byte *)&v, 4);
	if (v != 0xFFFFFFFF)
		return -1;

	ReadSRAM(fd, (byte *)buf, 4);
	if (strncmp(buf, GAMETYPE, 4))
		return -1;

	SeekSRAM(fd, 32, SEEK_SET);
	ReadSRAM(fd, (byte *)tag, 32);

	return 0;
}

int LoadTheGame(char *fn, int x, int y)
{
#ifndef USE_LZ77
	int fd;
	int i;
	long checksum, oldchecksum;
	objtype nullobj;

	fd = fn[7] - '0'; // filename was made like this - name[7] = which+'0';
	if ((fd >= NUM_SLOTS) || (fd < 0))
		return -1;

	SeekSRAM(fd, 64, SEEK_SET);

	checksum = 0;

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)&gamestate, sizeof(gamestate));
	checksum = DoChecksum((byte*)&gamestate,sizeof(gamestate),checksum);

	DiskFlopAnim(x,y);

#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		ReadSRAM(fd, (byte *)&LevelRatios[i], sizeof(LRstruct));
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	DiskFlopAnim(x,y);

	SetupGameLevel();

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap,sizeof(tilemap), checksum);

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	ReadSRAM(fd, (byte *)areaconnect, sizeof(areaconnect));
	ReadSRAM(fd, (byte *)areabyplayer, sizeof(areabyplayer));

	InitActorList();

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)player, sizeof(*player));

	while (1)
	{
		DiskFlopAnim(x,y);
		ReadSRAM(fd, (byte *)&nullobj, sizeof(nullobj));
		if (nullobj.active == ac_badobject)
			break;

		GetNewActor(); // don't copy over the links

		memcpy((void*)new, (void*)&nullobj, sizeof(nullobj));
	}

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);
	laststatobj = statobjlist + i;

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	DiskFlopAnim(x,y);

	ReadSRAM(fd, (byte *)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);

	ReadSRAM(fd, (byte *)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);

	ReadSRAM(fd, (byte *)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);

	ReadSRAM(fd, (byte *)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir,sizeof(pwalldir),checksum);

	ReadSRAM(fd, (byte *)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	DiskFlopAnim(x,y);

//	ReadSRAM(fd, (byte *)&Auto_Map, sizeof(automap_t));
//	checksum = DoChecksum((byte*)&Auto_Map, sizeof(automap_t),checksum);

	ReadSRAM(fd, (byte *)&oldchecksum, sizeof(oldchecksum));
	if (oldchecksum != checksum)
	{
		Message(STR_SAVECHT1"\n"
			 STR_SAVECHT2"\n"
			 STR_SAVECHT3"\n"
			 STR_SAVECHT4);

		IN_ClearKeysDown();
		IN_Ack();

		gamestate.score = 0;
		gamestate.lives = 1;
		gamestate.weapon =
		gamestate.chosenweapon =
		gamestate.bestweapon = wp_pistol;
		gamestate.ammo = 8;
	}

	// eventually replace this with loaded automap
	memset((void *)&Auto_Map, 0, sizeof(automap_t));
#else
	int fd;
	int i;
	long checksum, oldchecksum;
	objtype nullobj;
	byte *umem, *cmem;
	int csize;

#define ReadRAM(m, s) { memcpy(m, umem, s); umem += s; }

	VL_WaitVBL(5);

	fd = fn[7] - '0'; // filename was made like this - name[7] = which+'0';
	if ((fd >= NUM_SLOTS) || (fd < 0))
		return -1;

	SeekSRAM(fd, 64, SEEK_SET);
	ReadSRAM(fd, (byte*)&csize, 4);
	if (!csize || (csize > (16384-64-4)))
		return -1;

	cmem = cmpbuf;
	umem = gfxbuf;

	ReadSRAM(fd, cmem, csize);
	Decompress(cmem, umem, csize);

	checksum = 0;

	ReadRAM((byte *)&gamestate, sizeof(gamestate));
	checksum = DoChecksum((byte*)&gamestate,sizeof(gamestate),checksum);

#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		ReadRAM((byte *)&LevelRatios[i], sizeof(LRstruct));
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	SetupGameLevel();

	ReadRAM((byte *)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap,sizeof(tilemap), checksum);

	ReadRAM((byte *)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	ReadRAM((byte *)areaconnect, sizeof(areaconnect));
	ReadRAM((byte *)areabyplayer, sizeof(areabyplayer));

	InitActorList();

	ReadRAM((byte *)player, sizeof(*player));

	while (1)
	{
		ReadRAM((byte *)&nullobj, sizeof(nullobj));
		if (nullobj.active == ac_badobject)
			break;

		GetNewActor(); // don't copy over the links

		memcpy((void*)new, (void*)&nullobj, sizeof(nullobj));
	}

	ReadRAM((byte *)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);
	laststatobj = statobjlist + i;

	ReadRAM((byte *)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	ReadRAM((byte *)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);

	ReadRAM((byte *)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	ReadRAM((byte *)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);

	ReadRAM((byte *)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);

	ReadRAM((byte *)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);

	ReadRAM((byte *)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir,sizeof(pwalldir),checksum);

	ReadRAM((byte *)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	ReadRAM((byte *)&Auto_Map, sizeof(automap_t));
	checksum = DoChecksum((byte*)&Auto_Map, sizeof(automap_t),checksum);

	ReadRAM((byte *)&oldchecksum, sizeof(oldchecksum));
	if (oldchecksum != checksum)
	{
		Message(STR_SAVECHT1"\n"
			 STR_SAVECHT2"\n"
			 STR_SAVECHT3"\n"
			 STR_SAVECHT4);

		IN_ClearKeysDown();
		IN_Ack();

		gamestate.score = 0;
		gamestate.lives = 1;
		gamestate.weapon =
		gamestate.chosenweapon =
		gamestate.bestweapon = wp_pistol;
		gamestate.ammo = 8;
	}
#endif

	return 0;
}

/* ======================================================================== */

/*
=================
=
= MS_CheckParm
=
=================
*/

int MS_CheckParm(char *check)
{
	int i;
	char *parm;

	for (i = 1; i < _argc; i++) {
		parm = _argv[i];

		while (!isalpha(*parm))		// skip - / \ etc.. in front of parm
			if (!*parm++)
				break;		// hit end of string without an alphanum

		if (!stricmp(check, parm))
			return i;
	}
	return 0;
}

/* ======================================================================== */

/*
==================
=
= BuildTables
=
==================
*/

static const float radtoint = (float)FINEANGLES/2.0/PI;

void BuildTables()
{
	int i;
	float tang, angle, anglestep;
	fixed value;

/* calculate fine tangents */

	finetangent[0] = 0;
	for (i = 1; i < FINEANGLES/8; i++) {
		tang = tanf((float)i/radtoint);
		finetangent[i] = tang*TILEGLOBAL;
		finetangent[FINEANGLES/4-1-i] = TILEGLOBAL/tang;
	}

	/* fight off asymptotic behaviour at 90 degrees */
	finetangent[FINEANGLES/4-1] = finetangent[FINEANGLES/4-2]+1;

//
// costable overlays sintable with a quarter phase shift
// ANGLES is assumed to be divisable by four
//

	angle = 0.0;
	anglestep = PI/2.0/ANGLEQUAD;
	for (i = 0; i <= ANGLEQUAD; i++) {
		value = GLOBAL1*sinf(angle);

		sintable[i] =
		sintable[i+ANGLES] =
		sintable[ANGLES/2-i] = value;

		sintable[ANGLES-i] =
		sintable[ANGLES/2+i] = -value;

		angle += anglestep;
	}
}

/*
===================
=
= SetupWalls
=
= Map tile values to scaled pics
=
===================
*/

void SetupWalls()
{
	int i;

	for (i=1;i<MAXWALLTILES;i++)
	{
		horizwall[i]=(i-1)*2;
		vertwall[i]=(i-1)*2+1;
	}
}

void ShowViewSize(int width)
{
	int oldwidth,oldheight;

	oldwidth = viewwidthwin;
	oldheight = viewheightwin;

	viewwidthwin = width*16;
	viewheightwin = width*16/HEIGHTRATIO;
	DrawPlayBorder();

	viewheightwin = oldheight;
	viewwidthwin = oldwidth;
}

void NewViewSize(int width)
{
	if (width > 40)
		width = 40;
	if (width < 4)
		width = 4;

	width *= vwidth / 320;

	if ((width*16) > vwidth)
		width = vwidth / 16;

	if ((width*16/HEIGHTRATIO) > (vheight - 40*vheight/200))
		width = (vheight - 40*vheight/200)/8;

	viewwidthwin = width*16*320/vwidth;
	viewheightwin = width*16/HEIGHTRATIO*320/vwidth;
	viewsize = width*320/vwidth;

	viewwidth = width*16;
	viewheight = width*16/HEIGHTRATIO;

	centerx = viewwidth/2-1;
	shootdelta = viewwidth/10;

	yoffset = (vheight-STATUSLINES*vheight/200-viewheight)/2;
	xoffset = (vwidth-viewwidth)/2;

//
// calculate trace angles and projection constants
//
	CalcProjection(FOCALLENGTH);

}

//===========================================================================

#ifndef SPEARDEMO

#ifndef SPEAR
CP_iteminfo MusicItems={8,CTL_Y,6,0,32};
CP_itemtype MusicMenu[]=
{
	{1,"Get Them!",0},
	{1,"Searching",0},
	{1,"P.O.W.",0},
	{1,"Suspense",0},
	{1,"War March",0},
	{1,"Around The Corner!",0},

	{1,"Nazi Anthem",0},
	{1,"Lurking...",0},
	{1,"Going After Hitler",0},
	{1,"Pounding Headache",0},
	{1,"Into the Dungeons",0},
	{1,"Ultimate Conquest",0},

	{1,"Kill the S.O.B.",0},
	{1,"The Nazi Rap",0},
	{1,"Twelfth Hour",0},
	{1,"Zero Hour",0},
	{1,"Ultimate Conquest",0},
	{1,"Wolfpack",0}
};
#else
CP_iteminfo MusicItems={8,CTL_Y-20,9,0,32};
CP_itemtype MusicMenu[]=
{
	{1,"Funky Colonel Bill",0},
	{1,"Death To The Nazis",0},
	{1,"Tiptoeing Around",0},
	{1,"Is This THE END?",0},
	{1,"Evil Incarnate",0},
	{1,"Jazzin' Them Nazis",0},
	{1,"Puttin' It To The Enemy",0},
	{1,"The SS Gonna Get You",0},
	{1,"Towering Above",0}
};
#endif

static const int songs[] =
{
#ifndef SPEAR
	GETTHEM_MUS,
	SEARCHN_MUS,
	POW_MUS,
	SUSPENSE_MUS,
	WARMARCH_MUS,
	CORNER_MUS,

	NAZI_OMI_MUS,
	PREGNANT_MUS,
	GOINGAFT_MUS,
	HEADACHE_MUS,
	DUNGEON_MUS,
	ULTIMATE_MUS,

	INTROCW3_MUS,
	NAZI_RAP_MUS,
	TWELFTH_MUS,
	ZEROHOUR_MUS,
	ULTIMATE_MUS,
	PACMAN_MUS
#else
	XFUNKIE_MUS,             // 0
	XDEATH_MUS,              // 2
	XTIPTOE_MUS,             // 4
	XTHEEND_MUS,             // 7
	XEVIL_MUS,               // 17
	XJAZNAZI_MUS,            // 18
	XPUTIT_MUS,              // 21
	XGETYOU_MUS,             // 22
	XTOWER2_MUS              // 23
#endif
};

void DoJukebox()
{
	int which,lastsong=-1;
	unsigned start;

	IN_ClearKeysDown();
//	if (!AdLibPresent && !SoundBlasterPresent)
//		return;

	MenuFadeOut();

#if !defined(SPEAR) || !defined(UPLOAD)
	start = (US_RndT() % 3) * 6;
#else
	start = 0;
#endif

	CA_CacheGrChunk(STARTFONT+1);
#ifdef SPEAR
	CacheLump(BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#else
	CacheLump(CONTROLS_LUMP_START, CONTROLS_LUMP_END);
#endif
	CA_LoadAllSounds();

	fontnumber=1;
	ClearMScreen();
	VWB_DrawPic(112,184,C_MOUSELBACKPIC);
	DrawStripes(10);
	SETFONTCOLOR (TEXTCOLOR,BKGDCOLOR);

#ifndef SPEAR
	DrawWindow (CTL_X-2,CTL_Y-6,280,13*7,BKGDCOLOR);
#else
	DrawWindow (CTL_X-2,CTL_Y-26,280,13*10,BKGDCOLOR);
#endif

	DrawMenu (&MusicItems,&MusicMenu[start]);

	SETFONTCOLOR (READHCOLOR,BKGDCOLOR);
	PrintY = 15;
	WindowX = 0;
	WindowY = 320;
	US_CPrint("Robert's Jukebox");

	SETFONTCOLOR (TEXTCOLOR,BKGDCOLOR);
	VW_UpdateScreen();
	MenuFadeIn();

	do
	{
		which = HandleMenu(&MusicItems,&MusicMenu[start],NULL);
		if (which>=0)
		{
			if (lastsong >= 0)
				MusicMenu[start+lastsong].active = 1;

			StartCPMusic(songs[start + which]);
			MusicMenu[start+which].active = 2;
			DrawMenu (&MusicItems,&MusicMenu[start]);
			VW_UpdateScreen();
			lastsong = which;
		}
	} while(which>=0);

	MenuFadeOut();
	IN_ClearKeysDown();
#ifdef SPEAR
	UnCacheLump(BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#else
	UnCacheLump(CONTROLS_LUMP_START, CONTROLS_LUMP_END);
#endif
}
#endif

/* ======================================================================== */

/*
==========================
=
= ShutdownId
=
= Shuts down all ID_?? managers
=
==========================
*/

void ShutdownId()
{
	US_Shutdown();
	SD_Shutdown();
	IN_Shutdown();
	VW_Shutdown();
	CA_Shutdown();
	PM_Shutdown();
	//MM_Shutdown();
}

/*
=====================
=
= NewGame
=
= Set up new game to start from the beginning
=
=====================
*/

void NewGame(int difficulty, int episode)
{
	memset(&gamestate, 0, sizeof(gamestate));

	gamestate.difficulty = difficulty;
	gamestate.weapon = gamestate.bestweapon
		= gamestate.chosenweapon = wp_pistol;
	gamestate.health = 100;
	gamestate.ammo = STARTAMMO;
	gamestate.lives = 3;
	gamestate.nextextra = EXTRAPOINTS;
	gamestate.episode = episode;

	startgame = true;
}

/*
==========================
=
= InitGame
=
= Load a few things right away
=
==========================
*/

void InitGame()
{
	int i;

	//MM_Startup();
	PM_Startup();
	CA_Startup();
	VW_Startup();
	IN_Startup();
	SD_Startup();
	US_Startup();

//
// build some tables
//

	for (i = 0;i < MAPSIZE; i++)
	{
		farmapylookup[i] = i*64;
	}

	ReadConfig();

/* load in and lock down some basic chunks */

	CA_CacheGrChunk(STARTFONT);
	CA_CacheGrChunk(STARTTILE8);
	for (i = LATCHPICS_LUMP_START; i <= LATCHPICS_LUMP_END; i++)
		CA_CacheGrChunk(i);

	BuildTables();
	SetupWalls();

	NewViewSize(viewsize);


//
// initialize variables
//
	InitRedShifts();

	IN_CheckAck();
//
// HOLDING DOWN 'M' KEY?
//
#ifndef SPEARDEMO
	if (IN_KeyDown(sc_M))
		DoJukebox();
#endif
}

/*
=====================
=
= DemoLoop
=
=====================
*/

void DemoLoop()
{
	static int LastDemo;

	int i;
//
// main game cycle
//

	LastDemo = 0;

	StartCPMusic(INTROSONG);

	if (!NoWait)
		PG13();

	i = MS_CheckParm("playdemo");
	if (i && ((i+1) < _argc)) {
		i++;
		for (; i < _argc; i++) {
			if (_argv[i][0] == '-')
				break;
			IN_ClearKeysDown();
			//if (PlayDemoFromFile(_argv[i]))
			//	IN_UserInput(3 * 70);
		}
		VW_FadeOut();
	}

	if (MS_CheckParm("demotest")) {
	#ifndef SPEARDEMO
		for (i = 0; i < 4; i++)
			PlayDemo(i);
	#else
		PlayDemo(0);
	#endif
	}

	while (1)
	{
		while (!NoWait)
		{
//
// title page
//
			//MM_SortMem ();
#ifdef SPEAR
			CA_CacheGrChunk (TITLEPALETTE);

			CA_CacheGrChunk (TITLE1PIC);
			VWB_DrawPic (0,0,TITLE1PIC);
			CA_UnCacheGrChunk (TITLE1PIC);

			CA_CacheGrChunk (TITLE2PIC);
			VWB_DrawPic (0,80,TITLE2PIC);
			CA_UnCacheGrChunk(TITLE2PIC);
			VW_UpdateScreen();
			VL_FadeIn(0,255,grsegs[TITLEPALETTE],30);

			CA_UnCacheGrChunk (TITLEPALETTE);
#else
			VL_CacheScreen(TITLEPIC);
			VW_UpdateScreen ();
			VW_FadeIn();
#endif
			if (IN_UserInput(TickBase*15))
				break;
			VW_FadeOut();
//
// credits page
//
			VL_CacheScreen(CREDITSPIC);
			VW_UpdateScreen();
			VW_FadeIn ();
			if (IN_UserInput(TickBase*10))
				break;
			VW_FadeOut ();
//
// high scores
//
			DrawHighScores();
			VW_UpdateScreen();
			VW_FadeIn();

			if (IN_UserInput(TickBase*10))
				break;
//
// demo
//
			#ifndef SPEARDEMO
			PlayDemo(LastDemo++%4);
			#else
			PlayDemo(0);
			#endif

			if (playstate == ex_abort)
				break;
			StartCPMusic(INTROSONG);
		}

		VW_FadeOut();

		//if (IN_KeyDown(sc_Tab) && MS_CheckParm("debugmode"))
		//	RecordDemo();
		//else
			US_ControlPanel(0);

		if (startgame || loadedgame)
		{
			GameLoop();
			VW_FadeOut();
			StartCPMusic(INTROSONG);
		}
	}
}


/* ======================================================================== */


/*
==========================
=
= WolfMain
=
==========================
*/

int WolfMain(int argc, char *argv[])
{
    char temp[80];

	_argc = argc;
	_argv = argv;

	if (MS_CheckParm("version")) {
		sprintf(temp, "Game: %s\n", GAMENAME);
		Debug32xScreenPuts(temp);
		//Debug32xScreenPrintf("Game: %s\n", GAMENAME);
		Quit(NULL);
	}

    sprintf(temp, "Now Loading %s\n", GAMENAME);
    //Debug32xScreenPuts(temp);
	//Debug32xScreenPrintf("Now Loading %s\n", GAMENAME);

	CheckForEpisodes();

	InitGame();

	DemoLoop();

	Quit("Demo loop exited???");

	return 0;
}
